#!/bin/bash

shellfu import dist
shellfu import jat
shellfu import jat_dump

#
# Static file checking
#
# This library provides single function to help perform static file
# checking of installed RPM package.  By default,  several built-in
# checks are run.  If that is enough, all you need is call the
# function and it will take care of everything else.
#
# However, aside from this simple scenario, main strength of this
# function is that it already sorts files into lists in such a way
# that adding custom checks becomes easy.
#
# Another strength of this library is its simple but dynamic blacklisting
# so that your tests can be flexible but remain easy to understand and
# maintain.
#
#
# =head1 BUILT-IN CHECKS
#
# Built-in checks are few, to start with:
#
#  *  `syntax.python` -- syntax check for all possible Python scripts,
#
#  *  `syntax.bash` -- syntax check for all possible Bash scripts.
#
#  *  `conv.pam_d` -- check that files in /etc/pam.d/ are valid.
#
#  *  `conv.python.shebang` -- check for all possible Python scripts
#     that they use `#!/usr/bin/python`, not `#!/usr/bin/env`
#
#
# =head1 CUSTOM CHECKS
#
# Before understanding custom checks, we first need to look at
# way how file lists are organized.  Each check runs inside a directory
# organized like this:
#
#     .
#     ├── BLACKLIST
#     ├── all
#     │   ├── directories
#     │   ├── files
#     │   └── symlinks
#     ├── relics
#     ├── dist
#     │   └── pam.d
#     ├── ext
#     │   ├── NO.EXT
#     │   ├── bash
#     │   └── conf
#     ├── files.links
#     │   ├── _etc_foo_foo.conf -> /etc/foo/foo.conf
#     │   ├── _usr_share_lib64_foo_bar.so -> /usr/share/lib64/foo/bar.so
#     │   ├── _usr_share_lib64_foo_baz.so -> /usr/share/lib64/foo/baz.so
#     │   └── ...
#     ├── heur
#     │   ├── bash
#     │   └── python
#     ├── mime
#     │   ├── application
#     │   │   └── xml
#     │   └── text
#     │       ├── plain
#     │       └── x-shellscript
#     ├── rpmdata
#     │   ├── dump
#     │   ├── info
#     │   ├── list
#     │   ├── provides
#     │   ├── requires
#     │   └── scripts
#     └── tmp
#
# where "leaves" are individual path lists:
#
#  *  `all/..` - just based on basic UNIX type,
#  *  `BLACKLIST` - is generated by `rstatic__mkblacklist()`,
#  *  `ext/..` - lists per file extension,
#  *  `mime/..`  - lists per MIME type as detected by file(1) utility,
#  *  `heur/..` - lists based on heuristics wired in this module,
#  *  `dist/..` - lists based on conventions given by distribution.
#
# `files.links` is not a list but a "flat" directory with symlinks to
# all files in the package.  This is useful if you want to use a tool
# with recursive mode such as grep, eg. `grep BAD_STRING -R files.links`
# may have vastly better performance than calling grep for each file
# in the package.
#
# `tmp` is also not a list, it's a directory designated for custom checks
# to allow keeping ad-hoc data (see Guildelines section of this manual).
# `relics` is similar except that this directory will be kept and
# submitted using jat__submit() along with all lists.
#
# `rpmdata` is another 'special' case; files here are direct exports from
# RPM database; where each file represents output of rpm query, where
# the query option corresponds to file name under rpmdata.  For example,
# output of `rpm --info` is saved under `rpmdata/info`.  As an addition,
# rpmdata/dump.real contains report similar to rpmdata/dump, but reflecting
# real state of filesystem rather than RPM database.  See RPMDATA MIRROR
# REPORT section for more details.
#
# Now, custom checks are added by implementing a check function and
# specifying it (and one of above lists, if needed) when calling
# `rstatic__checks()`.
#
# There are two types of custom checks:
#
#
# =head2 Simple, typed check
#
# This is a simple function that runs on a particular path list as
# described above.  For each path, the check function is called with
# single argument of the path.
#
# The list can be relative path within the above structure, but also
# *another function*.  This helps make very specific checks by
# combining lists generated by this library and custom, package-specific
# filtering.
#
# The format can be one of:
#
#     SOME/LIST:SIMPLEFN
#     LISTFN:SIMPLEFN
#
# where LIST can be filename path relative to the directory described
# above.  Alternatively, you can write own function to produce the
# list (LISTFN); in that case you will use the second form.  SIMPLEFN
# is function  provided by you; it should accept path name of file being
# checked as single argument, and contain at least one JAT assert.
#
# For example, check 'all/files:foo' would call function `foo()` on each
# file (each line in `all/files` file).  Check 'heur/python:bar' would call
# `bar()` on each path from `heur/python` list.  Last, a check 'heya:baz'
# would run `baz()` with every line generated by calling `heya()`.
#
#
# =head2 Free form check
#
# Free-form check is just a user-defined function called in the
# path list directory described above.
#
# Two great advantages of *simple check* is that you don't have to care
# about the iteration code and it's easy to combine with any cached
# or custom-created lists.  Free-form check trades these feats for
# flexibility while still retaining access to cached lists.
#
# =head2 Guildelines
#
# When writing custom checks, try to follow these guidelines:
#
#  *  Do not change directory (if you absolutely have to, use pushd/popd).
#
#  *  Do not write anywhere except tmp/ sub-directory, created for you for
#     this purpose.
#
#  *  Prefix each assert with relevant filename (for simple checks, this
#     is passed as $1, for free-form, use absolute paths as read from
#     cached lists).
#
#
# =head1 BLACKLISTING
#
# If a function named `rstatic__mkblacklist()` is defined,
# it must print a blacklist in form of lines as:
#
#     TEST_NAME:PATH
#
# Where TEST_NAME is name of the built-in test or a custom check function
# (SIMPLEFN), from which the PATH should be excluded.
#
#
# =head1 RPMDATA MIRROR REPORT
#
# File rpmdata/dump.real is created by this module and contains data in
# similar format as rpm --dump (ie. rpmdata/dump file), but reflecting
# real stat  This file can be useful when debugging rpm verification
# failures.
#
# The report will never be exactly the same, though, for several reasons.
# First, some fields cannot be reproduced by this module (eg. isconfig) or
# not reliably so (eg. rdev, size for directories...).  Second, files that
# were deleted from filesystem will be simply skipped.
#
# Also note that rpm allows explicitly excluding some fields from the rpm
# verification using %noverify field.  That is, these fields could be
# different in rpmdata/dump.real, and rpm check could still pass.
#
#
# =head1 EXAMPLE
#
# Skipping some boilerplate, this is body of a simple example of
# this function in action:
#
#
#     usepy3() {
#         jat__eval -h "$(rstatic__hint) Python 3 is used"
#            -S 1 \
#             "head -1 $1 | grep '#/usr/bin/python3'"
#         ## the hint is not mandatory, but we have added one
#         ## and by use of rstatic__hint() we ensure
#         ## the hint contains file path and check name already
#         ## formatted to copy/paste into blacklist
#     }
#
#     modules() {
#         grep -w /usr/lib/foo/modules heur/python
#                                     #\
#         ## we'll re-use existing list heur/python and just
#         ## filter it according to our specific criterion
#     }
#
#     rstatic__mkblacklist() {
#         echo usepy3:/usr/share/doc/foo/bad_examples/obsolete.py
#                     #\
#         ## iow, this particular file will be skipped by
#         ## usepy3() test.  Few hints here:
#         ##  *  Instead of 'usepy3', we could also have used
#         ##     a custom built-in check name
#         ##  *  In case path  changes per distribution, we can
#         ##     pull it from rpm database
#     }
#
#     shellfu import rstatic
#
#     rstatic__checks -c modules:usepy3 foo
#
# Explanation:  The test is written for package 'foo'.  That is,
# `rpm -ql foo` will be used as basis for all file lists.  Built-in checks
# are called by default; we're topping this up with out custom "simple
# check".
#
# We could have used one of provided lists, eg. `heur/python:usepy3`
# ("things that look like Python") but we decided to focus only on
# particular subset of that.
#


#
# Blacklisting mode: 'filter', 'ignore' or 'verify' (default: 'filter')
#
# 'filter' is the usual mode when blacklist is working.  'ignore' will
# stop filtering, just as if blacklist was empty.
#
# 'verify' mode will try to detect obsolete blacklist items: it will
# run checks for all item, but will raise warning for items that would
# be filtered but did not really fail.
#
# It's recommended to run your tests with 'verify' once in a while, after
# a rebase or major change.
#
RSTATIC__BLACKLIST=${RSTATIC__BLACKLIST:-filter}

#
# Options for builtins
#
# Comma-separated list of K=V pairs where each KEY must
# start with builtin name, colon and option name.
#
#     BLTNNAME:OPTNAME=foo,BLTNNAME:OPTNAME=bar
#
# These values are available to the builtin via _rstatic__opt() function.
#
RSTATIC__BLTNOPTS=${RSTATIC__BLTNOPTS:-}

rstatic__hint() {
    #
    # Get hint for asserts inside custom check
    #
    # Assert comments are not mandatory but using this function you
    # this function will help you create hint:
    #
    #     CURRENT_CHECK:CURRENT_PATH
    #
    # which can be directly used in blacklist.
    #
    echo "$__RSTATIC__TEST_NAME:$__RSTATIC__ITEM"
}

rstatic__checks() {
    #
    # For a package, run built-in checks plus custom checks
    #
    # Usage:
    #
    #      rstatic__checks [-R] [-B] [-b BUILTIN].. [-c CUSTOM_CHECK].. PKG"
    #
    # For package PKG, make lists and run
    #
    #  *   basic RPM integrity check, unless suppressed by -R,
    #
    #  *   all built-in checks, unless suppresed by `-B` or specified by `-b`,
    #
    #  *   all custom checks, if specified by -c.
    #
    # Finally, it will collect all intermediate relics (file lists, RPM data
    # and files dropped by custom checks, if any) as `PKG-NVR-relics.tar.gz`
    # ans upload it using jat__submit for later reference.
    #
    # All necessary phases are created.
    #
    # Each CUSTOM_CHECK has to have one of this form:
    #
    #     SOME/LIST:SIMPLEFN
    #     LISTFN:SIMPLEFN
    #     FREEFN
    #
    # where SOME/LIST is a relative path to one of lists as described in
    # CUSTOM CHECKS section, LISTFN is name of a user-defined listing
    # function), SIMPLEFN is your simple check function name, and FREEFN
    # is name of your free-form check function.  Read CUSTOM CHECKS
    # section of this manual for details.
    #
    local Package           # package name to check
    local PackageNvr        # package NVR
    local lsthome           # list cache root
    local pktmp             # temporary directory for packaging relics
    local custs=()          # custom checks specified by `-c`
    local bltns=()          # builtin checks specified by `-b`
    local do_rpmv=true      # do `rpm -V` check?
    local do_bltns=true     # run builtins?
    local rpmv_tmp          # cache for `rpm -V` output

    while true; do case $1 in
        -R|--no-rpmv)    do_rpmv=false;  shift ;;
        -B|--no-builtin) do_bltns=false; shift ;;
        -c|--custom)     custs+=("$2");  shift 2 || { __rstatic__usage; return 2; } ;;
        -b|--builtin)    bltns+=("$2");  shift 2 || { __rstatic__usage; return 2; } ;;
        *) break ;;
    esac done
    Package=$1
    test -n "$2" && { __rstatic__usage; return 2; }
    test -n "$Package" || { __rstatic__usage; return 2; }

    PackageNvr=$(
        rpm --qf '%{NAME}-%{VERSION}-%{RELEASE}\n' -q "$Package" \
          | grep -vx 'package .* is not installed' \
          | sort \
          | uniq
    )

    test -n "$PackageNvr" || {
        jat__log_error \
            "package not installed: '$Package'"
        return 3
    }

    grep -qxF -e filter -e ignore -e verify <<<"$RSTATIC__BLACKLIST" || {
        jat__log_error \
            "invalid blacklisting mode: '$RSTATIC__BLACKLIST'"
        return 3
    }

    if $do_rpmv;
    then
        jat__pstarts -h "$PackageNvr: verify RPM integrity"
            rpmv_tmp=$(mktemp -t distribution_static.rpmv.XXXXXXXX)
            jat__cmd -h "do rpm --verify on $Package" \
                -o "$rpmv_tmp" \
                rpm -V "$Package"
            jat_dump__file --skip-empty "$rpmv_tmp"
        jat__pend
    fi

    lsthome=$(mktemp -d -t "rstatic__lsthome.XXXXXXXX")
    pushd "$lsthome" >/dev/null || {
        jat__log_error "could not move to temp"
        return 3
    }

        jat__pstarts -h "$PackageNvr: store rpm data"
            jat__cmd -h "Store RPM data" \
                __rstatic__make_rpmdata
            jat__cmd -h "Make rpm --dump \"mirror report\"" \
                __rstatic__make_rpmreal
        jat__pend

        jat__pstarts -h "$PackageNvr: make lists"
            jat__cmd -h "Make lists of our files and dirs" \
                __rstatic__make_lists
            jat__cmd -h "Remove duplicates (possible in case of multilib)" \
                __rstatic__squash_lists
            jat__cmd -h "Make MIME-type-based lists" \
                __rstatic__make_lists_mime
            jat__cmd -h "make extension-based lists" \
                __rstatic__make_lists_ext
            jat__cmd -h "make distribution-based lists" \
                __rstatic__make_lists_dist
            jat__cmd -h "make heuristic-based lists" \
                __rstatic__make_lists_heur
            jat__cmd -h "Make symlinks to our files in one big dir" \
                __rstatic__make_links
            touch BLACKLIST
            type -t "rstatic__mkblacklist" >/dev/null && {
                jat__cmd -h "Make blacklist" \
                    -S 0-255 \
                    -o BLACKLIST \
                    rstatic__mkblacklist
                jat_dump__file BLACKLIST
            }
            jat__log_info "RSTATIC__BLACKLIST='$RSTATIC__BLACKLIST'"
            case $RSTATIC__BLACKLIST in
                ignore) jat__log_warning "blacklist is ignored" ;;
                verify) jat__log_warning "blacklist is being verified" ;;
            esac

            jat__cmd -h "create relics/ for writing" \
                mkdir relics
            jat__cmd -h "create tmp/ for writing" \
                mkdir tmp

            {
                echo ITEMS LIST
                wc -l all/files
                wc -l all/directories
                wc -l all/symlinks
            } | column -t | jat_dump__pipe LIST_STATS
        jat__pend

        $do_bltns && __rstatic__run_builtins "${bltns[@]}"
        __rstatic__run_customs "${custs[@]}"

    popd || {
        jat__log_error "could not move out of tmp"
        return 3
    }

    pktmp=$(mktemp -d -t rstatic__pktmp.XXXXXXXX)
    jat__pstartc -C "$pktmp" -h "$PackageNvr: collect relics"
        mkdir "$PackageNvr-relics"
        mv "$lsthome"/* "$PackageNvr-relics"
        jat__cmd -h "remove tmp/" \
            rm -rf "$PackageNvr-relics/tmp"
        jat__cmd -h "create tarball of file lists and relics" \
            tar -czf "$PackageNvr-relics.tar.gz" "$PackageNvr-relics"
        jat__submit "$PackageNvr-relics.tar.gz"
    jat__pend
    rm -rf "$lsthome" "$pktmp"
}

_rstatic__mkblacklist() {
    #
    # Print blacklist for tests
    #
    # Each line has to have format:
    #
    #     TEST_NAME:PATH
    #
    # Where TEST_NAME is name of the test from which the PATH should be
    # excluded.   In case of custom checks, FN shall be used as TEST_NAME.
    #
    true "this is template for rstatic__mkblacklist"
}


#          #                    if you want something checked below this #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                       then sorry; you have been blacklisted #

#
# Path to file currently being run
#
__RSTATIC__ITEM=""

#
# Current test name
#
__RSTATIC__TEST_NAME=""

__rstatic__ck_conv_pam_d() {
    #
    # Check PAM module configs
    #
    local path=$1           # path to check file at
    local fail=false        # did check fail?
    local module            # each .so module seen in $path
    for module in $(grep -Eo 'pam_[[:alnum:]]+\.so' "$path" | sort | uniq);
    do
        jat__eval -h "$CkName:$path:$module" \
            "! yum provides '*/$module' | grep 'No Matches found'" \
         || fail=true
    done
    $fail && jat_dump__file -s "$path"
}

__rstatic__ck_conv_python_shebang() {
    #
    # Check Python binaries' shebang lines
    #
    # Based on this proposal:
    #
    #     https://fedoraproject.org/wiki/Features/SystemPythonExecutablesUseSystemPython
    #
    # The document is not complete on details but at some places the policy
    # is being implemented.  This test verifies that if shebang is present,
    # it starts with `#!/usr/bin/python`.
    #
    local path=$1       # path to check file at
    head -1 "$path" > tmp/line1
    grep -q '^#!' tmp/line1 || {
        debug "no shebang, skipping: $path"
        rm tmp/line1
        return 0
    }
    jat__eval -h "$CkName:$path" \
        "grep -qwEe '^#! */usr/bin/python[23]?' -e '^#! */usr/libexec/platform-python' tmp/line1" || {
        head "$path" | jat_dump__pipe FILE_HEAD
    }
    rm tmp/line1
}

__rstatic__rel_conv_python_shebang() {
    #
    # Decide if this test is relevant based on distribution
    #
    dist__test \
        id == rhel \
        ver vge "7.0"
}

__rstatic__ck_syntax_bash() {
    #
    # Check files for Bash syntax
    #
    local path=$1
    jat__cmd -h "$CkName:$path" -e err bash -n "$path"
    jat_dump__file -s err
}

__rstatic__ck_syntax_python_guessbin() {
    #
    # Guess which python binary is applicable
    #
    # First, consult RSTATIC__BLTNOPTS, then guess from path, next apply
    # heuristics on rpmdata/requires.   Finally scream a warning and use
    # hardcoded /usr/bin/python.
    #
    local reqs
    local it_was_2='echo /usr/bin/python2; return 0'
    local it_was_3='echo /usr/bin/python3; return 0'
    local it_was_P='echo /usr/libexec/platform-python; return 0'
    _rstatic__opt pybin && return 0
    case $path in
        /usr/lib/python2*) eval "$it_was_2" ;;
        /usr/lib/python3*) eval "$it_was_3" ;;
    esac
    reqs=$(
        grep python rpmdata/requires \
          | sed 's/[[:space:]][[:space:]]*/ /g; s/^[[:space:]]*//; s/[[:space:]]*$//'
    )
    if grep -qF 'python(abi) = 2.' <<<"$reqs"; then
        eval "$it_was_2"
    elif grep -qF 'python(abi) = 3.' <<<"$reqs"; then
        eval "$it_was_3"
    elif grep -q '/usr/bin/python2\>' <<<"$reqs"; then
        eval "$it_was_2"
    elif grep -q '/usr/bin/python3\>' <<<"$reqs"; then
        eval "$it_was_3"
    elif grep -q '/usr/libexec/platform-python\>' <<<"$reqs"; then
        eval "$it_was_P"
    else
        jat__log_warning \
            "$(rstatic__hint): could not guess python binary path, falling back to /usr/bin/python"
        jat__log_info \
            "hint: This probably means your package contains python files but does" \
            "      not specify \`python(abi)\` in Requires to help rstrap detect"   \
            "      which binary will be used to run the code.  If you are sure this"\
            "      is intended, you can mute this warning by providing correct"     \
            "      Python binary by adding 'syntax.python:pybin=/correct/binary'"   \
            "      to the RSTATIC__BLTNOPTS environment variable."                  \
            "      To disable the syntax check for one or more particular files,"   \
            "      see BLACKLISTING section of the rstatic documentation page."
        echo "/usr/bin/python"
        return 2
    fi
}

__rstatic__ck_syntax_python() {
    #
    # Check files for Python syntax
    #
    local path=$1       # path to check file at
    local pybin         # path to python binary
    pybin=$(__rstatic__ck_syntax_python_guessbin)
    echo "compile(open('$path', 'rU').read(), '', 'exec')" \
      | jat__cmd -h "$CkName:$path" \
            -e err \
            "$pybin" -
    jat_dump__file -s err
}

__rstatic__rel_rpmdata_fedorahosted() {
    #
    # True if freeform.rpmdata.fedorahosted is relevant here
    #
    dist__test \
        id == rhel \
        ver vge "7.5" \
     && return 0
    dist__test \
        id == fedora \
        ver vge "25" \
     && return 0
    return 1
}

__rstatic__ck_rpmdata_fedorahosted() {
    #
    # fedorahosted.org has beeen retired
    #
    # https://fedoraproject.org/wiki/Infrastructure/Fedorahosted-retirement
    #
    jat__cmd -h "$(rstatic__hint): fedorahosted.org not mentioned in rpm info" \
        -S 1 \
        grep -qFw fedorahosted.org rpmdata/info
    test "$(jat__stat pfailc)" -gt 0 \
     && jat_dump__file rpmdata/info
}

__rstatic__make_links() {
    #
    # Create set of symlinks for faster grepping (grep -R)
    #
    local path          # each path
    local safename      # "safe" version of path
    mkdir files.links
    while read -r path;
    do
        safename="${path////_}"         # '/' => '_'
        safename="${safename// /_}"     # ' ' => '_'
        ln -s "$path" "files.links/$safename" || return 1
    done <all/files
}

__rstatic__make_rpmdata() {
    #
    # Make RPM info in rpm directory
    #
    local arg
    mkdir -p rpmdata
    for arg in $(__rstatic__rpmargs);
    do
        rpm -q --"$arg" "$Package" >"rpmdata/$arg"
    done
}

__rstatic__make_rpmreal() {
    #
    # Store real state of info tracked in rpm database
    #
    # Generate output in same format as rpm --dump, but reflecting
    # real state of files.
    #
    local path          # path as in rpm --dump (but reality)
    local size          # ^^
    local mtime         # ^^
    local digest        # ^^
    local mode          # ^^
    local owner         # ^^
    local group         # ^^
    local isconfig='~'  # ^^
    local isdoc='~'     # ^^
    local rdev='~'      # ^^
    local symlink       # ^^
    local stats         # some of above collected by stat
    local hmode         # full mode in hex format (to make $mode)
    cut -d' ' -f1 rpmdata/dump \
      | while read -r path; do
            symlink=X
            digest=0000000000000000000000000000000000000000000000000000000000000000
            test -e "$path" || continue
            stats=$(stat -c "size=%s; mtime=%Y; hmode=%f; owner=%U; group=%G" "$path")
            eval "$stats"   # safe because none of above ^^ can include funny chars
            test -d "$path" && size='~'
            if test -L "$path"; then
                symlink=$(readlink "$path")
            elif test -f "$path"; then 
                digest=$(sha256sum "$path")
                digest=${digest%% *}
            fi
            mode=$(printf '0%o\n' "0x$hmode" | sed 's/^00*/0/')     # ..is how --dump formats it
            echo "$path $size $mtime $digest $mode $owner $group $isconfig $isdoc $rdev $symlink"
        done >rpmdata/dump.real
}

__rstatic__rpmargs() {
    #
    # List valid /usr/bin/rpm args
    #
    {
        echo info
        echo changelog
        echo list
        echo dump
        echo configfiles
        echo docfiles
    #   echo conflicts
    #   echo enhances
    #   echo obsoletes
        echo provides
    #   echo recommends
        echo requires
    #   echo suggests
    #   echo supplements
        echo scripts
        echo triggers
        echo triggerscripts
    }
    #FIXME: enable also more recent args
}

__rstatic__make_lists() {
    #
    # Make lists of "our" files/dirs
    #
    local path          # each path
    mkdir -p all
    touch all/symlinks all/directories all/files
    rpm -ql "$Package" \
      | sort \
      | while read -r path;
        do
            test -L "$path" && echo "$path" >> all/symlinks && continue
            test -d "$path" && echo "$path" >> all/directories
            test -f "$path" && echo "$path" >> all/files
        done
    true
}

__rstatic__make_lists_ext() {
    #
    # Make lists based on file extension
    #
    local fpath     # path from 'files' list
    local fname     #  ^^  basename
    local fnext     #       ^^  extension
    local fncore    #       ^^  name "core" (extension dropped)
    mkdir -p ext
    while read -r fpath;                # /path/to/file.xml || /to/README
    do
        fname="${fpath##*/}"            # 'file.xml' || 'README'
        fncore="${fname%.*}"            # 'file'     || 'README'
        fnext="${fname#$fncore}"        # '.xml'     || ''
        test -n "$fnext" || {
            echo "$fpath" >> "ext/NO.EXT"
            continue
        }
        fnext="${fnext#.}"              # 'xml'      || ''
        echo "$fpath" >> "ext/$fnext"
    done <all/files
}

__rstatic__make_lists_mime() {
    #
    # Make lists based on MIME type
    #
    local line          # each line from file(1) utility
    local fpath         #       ^^ only file path
    local mtype         #       ^^ only MIME type
    local mtype_a       #                ^^ only "left" part
    local mtype_b       #                ^^ only "right" part
    mkdir -p mime
    file --mime-type -f all/files \
      | while read -r line;                    # /path/to/file:    text/html
        do
            mtype="${line#*:}"              # '    text/html'
            mtype="${mtype// /}"            # 'text/html'
            fpath="${line%:*}"              # '/path/to/file'
            mtype_a="${mtype%/*}"           # 'text'
            mtype_b="${mtype#*/}"           # 'html'
            test -n "$mtype_a" -a -n "$mtype_b" || {
                jat__log_warning "weird MIME type; ignoring: $mtype"
                continue
            }
            mkdir -p "mime/$mtype_a"
            echo "$fpath" >> "mime/$mtype_a/$mtype_b"
        done
}

__rstatic__make_lists_dist() {
    #
    # Make lists based on distribution conventions
    #
    mkdir -p dist
    grep ^/etc/pam.d ext/NO.EXT \
      > dist/pam.d
    true
}

__rstatic__make_lists_heur() {
    #
    # Make lists based on heuristics
    #
    mkdir -p heur

    cat mime/text/x-shellscript \
        ext/sh \
        ext/bash \
        2>/dev/null \
      | sort \
      | uniq \
      > heur/bash

    cat mime/text/x-python \
        ext/py \
        2>/dev/null \
      | sort \
      | uniq \
      > heur/python
}


__rstatic__ls_builtins() {
    #
    # List all builtin tests to consider
    #
    # If any arguments are seen, list them; otherwise list all known
    # (hardcoded) builtins.
    #
    test -n "$1" && {
        printf '%s\n' "$@"
    }
    echo conv.python.shebang
    echo syntax.bash
    echo syntax.python
    echo conv.pam_d
    echo freeform.rpmdata.fedorahosted
}

_rstatic__opt() {
    #
    # Value of option $1 of current builtin check from $RSTATIC__BLTNOPTS
    #
    # Return 1 if value was not specified, 0 otherwise.  (Value can be
    # specified as empty.)
    #
    local wkey=$1
    local match
    match=$(
        #shellcheck disable=SC2001
        sed 's/,/\n/' <<<"$RSTATIC__BLTNOPTS" \
          | grep -e "^$CkName:$wkey="
    )
    test -n "$match" || return 1
    echo "${match#*=}"
}

__rstatic__run_builtins() {
    #
    # Run built-in check $1
    #
    local CkName        # each check name
    local fn            # each check function
    local list          # each check input list
    local relevant      # each check relevancy function name
    for CkName in $(__rstatic__ls_builtins "$@");
    do
        relevant=true
        case $CkName in
            conv.pam_d)
                list=dist/pam.d
                fn=__rstatic__ck_conv_pam_d
                ;;
            conv.python.shebang)
                list=heur/python
                fn=__rstatic__ck_conv_python_shebang
                relevant=__rstatic__rel_conv_python_shebang
                ;;
            syntax.bash)
                list=heur/bash
                fn=__rstatic__ck_syntax_bash
                ;;
            syntax.python)
                list=heur/python
                fn=__rstatic__ck_syntax_python
                ;;
            freeform.rpmdata.fedorahosted)
                relevant=__rstatic__rel_rpmdata_fedorahosted
                fn=__rstatic__ck_rpmdata_fedorahosted
                ;;
            *)
                jat__log_error \
                    "unknown check: $CkName"
                return 3
                ;;
        esac
        case $CkName in
            freeform.*)
                jat__pstartt -h "$PackageNvr: built-in free-form check: $CkName"
                    if ! $relevant; then
                        jat__log_info "skipping test not applicable for this distribution"
                    else
                        __rstatic__run_ffck $fn "$CkName"
                    fi
                jat__pend
                ;;
            *)
                jat__pstartt -h "$PackageNvr: built-in file-based check: $CkName"
                    if ! $relevant; then
                        jat__log_info "skipping test not applicable for this distribution"
                    elif test "$(wc -l <$list)" -eq 0; then
                        jat__log_info "no applicable items"
                    else
                        <$list __rstatic__run_ck $fn "$CkName"
                    fi
                jat__pend
                ;;
        esac
    done
}

__rstatic__run_customs() {
    #
    # Run custom check $1
    #
    local ckdef     # each check definition expression
    local list      # list specification (relpath in list cache)
    local ckfn      # custom check function
    local lsfn      # custom listing function
    for ckdef in "$@";
    do
        case $ckdef in

            /:|:*|*:|/:*|*:/)
                jat__log_error \
                    "invalid custom check definition: '$ckdef'"
                ;;

            # simple check with list
            #
            */*:*)
                list=${ckdef%:*}
                ckfn=${ckdef#*:}
                jat__pstartt -h "$PackageNvr: custom simple check: $ckdef"
                    type -t "$ckfn" >/dev/null || {
                        jat__log_error \
                            "no such check function: '$ckfn'"
                        continue
                    }
                    test -f "$list" || {
                        jat__log_error \
                            "no such list: '$list'"
                        return 3
                    }
                    case $(wc -l <"$list") in
                        0)  jat__log_info "no applicable items" ;;
                        *)  <"$list" __rstatic__run_ck "$ckfn" "$ckfn" ;;
                    esac
                jat__pend
                ;;

            # simple check with function
            #
            *:*)
                ckfn=${ckdef#*:}
                lsfn=${ckdef%:*}
                jat__pstartt -h "$PackageNvr: custom simple check: $ckdef"
                    type -t "$ckfn" >/dev/null || {
                        jat__log_error \
                            "no such check function: '$ckfn'"
                        continue
                    }
                    type -t "$lsfn" >/dev/null || {
                        jat__log_error \
                            "no such listing function: '$lsfn'"
                        continue
                    }
                    ( "$lsfn" ) \
                      | __rstatic__run_ck "$ckfn" "$ckfn"
                jat__pend
                ;;

            # free-form check
            #
            *)
                jat__pstartt -h "$PackageNvr: custom free-form check: $ckdef"
                    type -t "$ckdef" >/dev/null || {
                        jat__log_error \
                            "check function not defined: '$ckfn'"
                        continue
                    }
                    ( "$ckdef" )
                jat__pend
                ;;

        esac
    done
}

__rstatic__run_ffck() {
    #
    # Run free-form test fn $1 under name prefix $2
    #
    local test_fn=$1        # test function name
    local test_name=$2      # test name
    case $RSTATIC__BLACKLIST in
        filter)
            grep -qxFe "$test_name:" BLACKLIST && {
                jat__log_info "skipping blacklisted test"
                return 0
            }
            ( __RSTATIC__ITEM="" \
              __RSTATIC__TEST_NAME="$test_name" \
              "$test_fn" )
        ;;
        ignore)
            ( __RSTATIC__ITEM="" \
              __RSTATIC__TEST_NAME="$test_name" \
              "$test_fn" )
        ;;
        verify)
            if grep -qxFe "$test_name:" BLACKLIST;
            then
                ( __RSTATIC__ITEM="" \
                  __RSTATIC__TEST_NAME="$test_name" \
                    "$test_fn" )
            test "$(jat__stat pfailc)" == 0 \
             && __rstatic__report_stale "$test_name:"
            else
                ( __RSTATIC__ITEM="" \
                  __RSTATIC__TEST_NAME="$test_name" \
                  "$test_fn" )
            fi
        ;;
    esac
}

__rstatic__squash_lists() {
    #
    # Remove duplicates from basic lists
    #
    # Note that due to how multilib works, file lists may contain
    # duplicate items (`rpm -ql a-multilib-package`).
    #
    < all/symlinks    sort | uniq > all/symlinks.uniq
    < all/directories sort | uniq > all/directories.uniq
    < all/files       sort | uniq > all/files.uniq
    mv all/symlinks{.uniq,}
    mv all/directories{.uniq,}
    mv all/files{.uniq,}
}

__rstatic__usage() {
    #
    # Bad usage
    #
    jat__log_error \
        "usage: rstatic__checks [-R] [-B] [-b BUILTIN]... [-c CUSTOM_CHECK]... [PKG]"
}

__rstatic__report_stale() {
    #
    # Report possibly stale keys $@
    #
    test -n "$1" || return 0
    jat__log_error \
        "following items on blacklist did not trigger fail:" \
        "$@"
}

__rstatic__run_ck() {
    #
    # Run test fn $1 under name prefix $2 with all paths from stdin
    #
    local test_fn=$1        # test function name
    local test_name=$2      # test name
    local item              # each path to run check function against
    local errs_pre          # number of errors before a check
    local errs_post         #              ^.. after a check
    local blist_mode        # blacklist mode
    local blist_key         # each processed blacklist item
    local blist_stale=()    # stale blacklist items (collected in verify mode)
    blist_mode=$RSTATIC__BLACKLIST
    while read -r item;
    do
        blist_key=$test_name:$item
        case $blist_mode in
            filter)
                grep -q "^$blist_key\\b" BLACKLIST && continue
                ( __RSTATIC__ITEM="$item" \
                  __RSTATIC__TEST_NAME="$test_name" \
                  "$test_fn" "$item" )
            ;;
            ignore)
                ( __RSTATIC__ITEM="$item" \
                  __RSTATIC__TEST_NAME="$test_name" \
                  "$test_fn" "$item" )
            ;;
            verify)
                if grep -q "^$blist_key\\b" BLACKLIST;
                then
                    errs_pre=$(jat__stat pfailc)
                    ( __RSTATIC__ITEM="$item" \
                      __RSTATIC__TEST_NAME="$test_name" \
                        "$test_fn" "$item" )
                    errs_post=$(jat__stat pfailc)
                    test "$errs_post" -eq "$errs_pre" \
                     && test "$blist_mode" = "verify" \
                     && blist_stale=("$blist_key")
                else
                    ( __RSTATIC__ITEM="$item" \
                      __RSTATIC__TEST_NAME="$test_name" \
                      "$test_fn" "$item" )
                fi
            ;;
        esac
    done
    __rstatic__report_stale "${blist_stale[@]}"
}

#shellfu module-version=0.0.12
